/*
 * AIEngine a new generation network intrusion detection system.
 *
 * Copyright (C) 2013-2023  Luis Campo Giralte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com> 
 *
 */
#ifndef SRC_PROTOCOLS_SNMP_SNMPPROTOCOL_H_
#define SRC_PROTOCOLS_SNMP_SNMPPROTOCOL_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Protocol.h"
#include <arpa/inet.h>

namespace aiengine {

// micro snmp ber header
struct snmp_header { 
	uint8_t 	code;
	uint8_t 	length;
	uint8_t 	type;
	uint8_t 	version_length;
	uint8_t 	data[0]; // snmp data 
} __attribute__((packed));

enum snmp_ber_types {
	SNMP_GET_REQ = 0xA0,
	SNMP_GET_NEXT_REQ = 0xA1,
	SNMP_GET_RES = 0xA2,
	SNMP_SET_REQ = 0xA3
};

class SNMPProtocol: public Protocol {
public:
	explicit SNMPProtocol():
		Protocol("SNMP", IPPROTO_UDP) {}
    	virtual ~SNMPProtocol();

	static constexpr int header_size = sizeof(struct snmp_header);

	uint16_t getId() const override { return 0x0000; }
	int getHeaderSize() const override { return header_size; }

	// Condition for say that a packet is snmp 
	bool check(const Packet &packet) override;	
        void processFlow(Flow *flow) override;
        bool processPacket(Packet &packet) override { return true; } 

	void statistics(std::basic_ostream<char> &out, int level, int32_t limit) const override;
	void statistics(Json &out, int level) const override;

	void releaseCache() override {} // No need to free cache

	void setHeader(const uint8_t *raw_packet) override { 

		header_ = reinterpret_cast <const snmp_header*> (raw_packet);
	}

	uint8_t getLength() const { return header_->length; }
	uint8_t getVersionLength() const { return header_->version_length; }

	int64_t getCurrentUseMemory() const override { return sizeof(SNMPProtocol); }
	int64_t getAllocatedMemory() const override { return sizeof(SNMPProtocol); }
	int64_t getTotalAllocatedMemory() const override { return sizeof(SNMPProtocol); }
	int64_t getAllocatedMemory(int value) const { return sizeof(SNMPProtocol); }

        void setDynamicAllocatedMemory(bool value) override {}
        bool isDynamicAllocatedMemory() const override { return false; }

	uint32_t getTotalEvents() const override { return total_events_; }

	CounterMap getCounters() const override; 
	void resetCounters() override;

	void setAnomalyManager(SharedPointer<AnomalyManager> amng) override { anomaly_ = amng; }
private:
	const snmp_header *header_ = nullptr;
	uint32_t total_events_ = 0;
	uint32_t total_snmp_get_requests_ = 0;
	uint32_t total_snmp_get_next_requests_ = 0;
	uint32_t total_snmp_get_responses_ = 0;
	uint32_t total_snmp_set_requests_ = 0;
	SharedPointer<AnomalyManager> anomaly_ = nullptr;
};

typedef std::shared_ptr<SNMPProtocol> SNMPProtocolPtr;

} // namespace aiengine

#endif  // SRC_PROTOCOLS_SNMP_SNMPPROTOCOL_H_
